import numpy as np
import pandas as pd
import networkx as nx

from typing import Optional, List
from tqdm import tqdm
from maxcutpy.abstractmaxcut import AbstractMaxCut

class RandomMaxCut(AbstractMaxCut):

    def batch_split(self):
        if self.batches_split:
            return self.best_cut_vector
        else:
            p = 0.5 * np.ones((self.matrix.shape[0]))
            cut_vec = self.generate_cut_vectors(1, self.matrix.shape[0], p)
            self.best_cut_score = self.score_cut_vectors(cut_vec)[0]

            self.best_cut_vector = cut_vec[0]

            self.batches_split = True
            return self.best_cut_vector



import numpy as np
import pandas as pd
import networkx as nx

from typing import Optional, List

from abc import ABC, abstractmethod


class AbstractMaxCut(ABC):

    def __init__(self, seed: Optional[int] = 12345, matrix: Optional[np.array] = None, graph: Optional[nx.Graph] = None):
        self.seed = seed
        self.generator = np.random.default_rng(self.seed)

        if matrix is not None:
            self.matrix = matrix
        elif graph is not None:
            self.matrix = nx.to_numpy_matrix(graph)
        else:
            self.matrix = np.array([[0, 1, 1], [1, 0, 2], [1, 2, 0]])

        self.batches_split = False
        self.best_cut_vector = None
        self.best_cut_score = None

    def score_cut_vectors(self, cut_vectors: np.array) -> List[np.float64]:
        """A function to calculate the score of a set of sample cut vectors

        Args:
            cut_vectors:    A numpy matrix of the cut vectors.
                Must be of shape (N, n) where n is the number of vertices and
                N is the number of sample cut vectors

        Returns:
            A List of np.float64 scores, N in length, which is the score of each
            cut vector

        """

        ##Needs to be replaced by a better, vectorized (or parallel) method
        scores = []
        for i in range(cut_vectors.shape[0]):
            cut_0 = np.argwhere(cut_vectors[i,:] == 0).reshape((1,-1))[0]
            cut_1 = np.argwhere(cut_vectors[i,:] == 1).reshape((1,-1))[0]

            pairs = np.array(np.meshgrid(cut_0, cut_1)).T.reshape(-1, 2)

            score_i = sum([self.matrix[x, y] for x, y in pairs])
            scores.append(score_i)

        return scores

    def generate_cut_vectors(self, N: int, n: int, p: List[np.float64]) -> np.array:
        """A function to generate a sample set of cut vectors

        Args:
            N: The integer number of samples to generate
            n: The integer number of vertices in each cut
            p: A list of float probabilities used in generating the cut vectors

        Returns:
            A numpy matrix of shape (N, n) where each row is a cut vector

        """
        assert len(p) == n

        cut_vectors = self.generator.binomial(n=1, p=p, size=(N, n))

        return cut_vectors

    @abstractmethod
    def batch_split(self) -> np.array:
        """A function to split the batches.

        This function is an abstract method in the class AbstractMaxCut.
        Should rely on internal attributes, and take no input parameters.

        Should return a single cut vector, which is a numpy array of shape (1, n), where
        n is the number of vertices in the matrix
        """
        pass



import numpy as np
import pandas as pd
import networkx as nx

from typing import Optional, List
from tqdm import tqdm
from maxcutpy.abstractmaxcut import AbstractMaxCut

class CrossEntropyMaxCut(AbstractMaxCut):

    def __init__(self, seed: Optional[int] = 12345, matrix: Optional[np.array] = None, p_cutoff: Optional[np.float64] = 0.1, N: Optional[int] = 1000, convergence_threshold: Optional[np.float64] = 0.0001, max_iters: Optional[int] = 1000):

        super().__init__(seed=seed, matrix=matrix)

        self.p_cutoff = p_cutoff
        self.N = N
        self.convergence_threshold = convergence_threshold
        self.max_iters = max_iters


    def batch_split(self):
        if self.batches_split:
            return self.best_cut_vector
        else:
            p = 0.5 * np.ones((self.matrix.shape[0]))
            t = 1

            for t in tqdm(range(self.max_iters)):
                cut_vectors = self.generate_cut_vectors(self.N, self.matrix.shape[0], p)
                scores = np.array(self.score_cut_vectors(cut_vectors))

                gamma = np.quantile(scores, 1 - self.p_cutoff, interpolation="nearest")

                thresh_scores = (scores >= gamma).astype(int)
                total_acheived = thresh_scores.sum()

                new_p = np.sum((cut_vectors * thresh_scores.reshape(-1, 1)), axis=0) / total_acheived

                if np.abs(new_p - p).mean() < self.convergence_threshold:
                    print("Results Converged!")
                    break

                p = new_p
                t += 1
            if t == self.max_iters - 1:
                print("Results failed to converge!")
                print(f"Current score is {max(scores)}")
            else:
                print(f"Converged at {t} iterations!")
                print(f"Found best score of {max(scores)}")

            self.best_cut_vector = cut_vectors[np.argmax(scores), :]
            self.best_cut_score = max(scores)
            self.batches_split = True
            return self.best_cut_vector


if __name__ == "__main__":
    maxcut = CEMaxCut(N=1000)

    print(maxcut.batch_split())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['maxcutpy']

package_data = \
{'': ['*']}

install_requires = \
['networkx>=2.5.1,<3.0.0',
 'numpy>=1.20.3,<2.0.0',
 'pandas>=1.2.4,<2.0.0',
 'tqdm>=4.61.0,<5.0.0']

setup_kwargs = {
    'name': 'maxcutpy',
    'version': '0.1.1',
    'description': 'A Python Implementation of Graph Max Cut Solutions',
    'long_description': '# maxcutpy\n A Python Implementation of Graph Max Cut Solutions\n\n## Problem Statement\nThe goal of this script is to provide simple python libraries for solving the [Maximum cut problem](https://en.wikipedia.org/wiki/Maximum_cut).\n\n## Current Solvers\n\n### Universal\nThe Structure of all Solvers is the same, from an API point of view.\nThis allows for quick testing of multiple different solvers, to find the best one.\n\nIn general, the flow will look like this:\n\n    from maxcutpy import RandomMaxCut\n    import numpy as np\n\n    matrix = np.array([[0,1,1],[1,0,5],[1,5,0]])\n    random_max_cut = RandomMaxCut(matrix=matrix, seed=12345)\n\n    best_cut_vector = random_cut.batch_split()\n\n`best_cut_vector` will be a numpy array of shape `(1,n)` where n is the number of rows in the input matrix. Each number inside of `best_cut_vector` will be an integer 0 or 1, depending on if it belongs to the 0th Slice or the 1st Slice.\n\nYou can then generally check the score of this `best_cut_vector` by using the same object again:\n\n    best_cut_score = random_cut.best_cut_score\n\nAt this time, a single Solver Object is Single-Use only, meaning it gets created for a specific matrix, and provides a batch split for this matrix only. Once the result is calculated, it will be stored, and repeated calls to `batch_split()` will only return the cached result.\n\nFor now, the matrix must be in the form of a numpy adjacency matrix, but support will be added soon for networkx Graph objects, as well as helper functions to transform between the two.\n\n### RandomMaxCut\nThis solve provides a method to compare against, as all this Solver does is select a random set of vertices to cut, giving each vertex a 50% chance of occurring in a different slice of the graph.\n\n    from maxcutpy import RandomMaxCut\n    import numpy as np\n\n    matrix = np.array([[0,1,1],[1,0,5],[1,5,0]])\n    random_max_cut = RandomMaxCut(matrix=matrix, seed=12345)\n\n    best_cut_vector = random_cut.batch_split()\n\n`best_cut_vector` will be a numpy array of shape `(1,n)` where each entry is a random integer 0 or 1, with 50% probability for either event. This means this could lead to the case where no cuts are made at all, yielding a score of 0.\n',
    'author': 'trevorWieland',
    'author_email': 'trevor_wieland@mac.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/trevorWieland/maxcutpy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
